﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GFramework.Base;
using Microsoft.Xna.Framework;

namespace GFramework.Camera
{
    public class Camera : IDirected,ICamera
    {
        public Guid ID { get; private set; }

        private Vector3 _position;
        public Vector3 Position
        {
            get { return _position; }
            set
            {
                Vector3 oldPos = _position;
                _position = value;
                if (PositionChanged != null)
                {
                    PositionChangedEventArgs args = new PositionChangedEventArgs(oldPos, _position);
                    PositionChanged(this, args);
                }
            }
        }
        public event EventHandler<PositionChangedEventArgs> PositionChanged;

        
        private Quaternion _orientation;

        public Vector3 Direction
        {
            get
            {
                Vector3 dir = Vector3.Zero;
                dir.X = -2.0f *
                    ((_orientation.X * _orientation.Z) + (_orientation.W * _orientation.Y));
                dir.Y = 2.0f *
                    ((_orientation.W * _orientation.X) - (_orientation.Y * _orientation.Z));
                dir.Z =
                    ((_orientation.X * _orientation.X) + (_orientation.Y * _orientation.Y)) -
                    ((_orientation.Z * _orientation.Z) + (_orientation.W * _orientation.W));
                Vector3.Normalize(ref dir, out dir);
                return dir;
            }
            set
            {
                Vector3 oldDir = Direction;
                Vector3 target = Position + value;
                Matrix temp = Matrix.CreateLookAt(Position, target, Vector3.Up);

                //invert the matrix, since we're determining the
                //orientation from the rotation matrix in RH coords
                temp = Matrix.Invert(temp);
                _orientation = Quaternion.CreateFromRotationMatrix(temp);
                
                if (DirectionChanged != null)
                {
                    PositionChangedEventArgs args = new PositionChangedEventArgs(oldDir, Direction);
                    DirectionChanged(this, args);
                }
            }
        }
        public event EventHandler<PositionChangedEventArgs> DirectionChanged;



        #region Параметры камеры
        public BoundingFrustum Frustum { get {return new BoundingFrustum(View * Projection);} }
        public Matrix View { get; private set; }
        public Matrix Projection { get; private set; }
        #endregion

        #region Вращения
        public void RotateClockwise(float angle)
        {
            Vector3 oldDirection = Direction;
            //rotate the aspect by the angle 
            _orientation = _orientation *
                Quaternion.CreateFromAxisAngle(Vector3.Forward, angle);

            //normalize to reduce errors
            Quaternion.Normalize(ref _orientation, out _orientation);
            if (DirectionChanged != null)
            {
                PositionChangedEventArgs args = new PositionChangedEventArgs(oldDirection, Direction);
                DirectionChanged(this, args);
            }
        }

        public void Rotate(Vector3 axisRotation)
        {
            Vector3 oldDirection = Direction;
            Quaternion quat = Quaternion.CreateFromYawPitchRoll(axisRotation.Y, axisRotation.X, axisRotation.Z);
            _orientation = _orientation * quat;
            _orientation.Normalize();
            if (DirectionChanged != null)
            {
                PositionChangedEventArgs args = new PositionChangedEventArgs(oldDirection, Direction);
                DirectionChanged(this, args);
            }
        }

        public void Rotate(float xrotation, float yrotation, float zrotation)
        {
            Rotate(new Vector3(xrotation, yrotation, zrotation));
        }

        public void RotateUp(float angle)
        {
            Vector3 oldDirection = Direction;
            //rotate the aspect by the angle 
            _orientation = _orientation *
                Quaternion.CreateFromAxisAngle(Vector3.Right, angle);

            //normalize to reduce errors
            Quaternion.Normalize(ref _orientation, out _orientation);
            if (DirectionChanged != null)
            {
                PositionChangedEventArgs args = new PositionChangedEventArgs(oldDirection, Direction);
                DirectionChanged(this, args);
            }
        }
        #endregion

        #region Движения
        public void StrafeLeft(float distance)
        {
            Position -= Right * distance;
        }

        public void StrafeRight(float distance)
        {
            Position += Right * distance;
        }

        public void MoveForward(float distance)
        {
            Position += Direction * distance;
        }

        public void MoveBackward(float distance)
        {
            Position -= Direction * distance;
        }
        #endregion

        #region Основные направления
        public Vector3 Up
        {
            get
            {
                Vector3 up = Vector3.Zero;
                up.X = 2.0f *
                    ((_orientation.X * _orientation.Y) - (_orientation.Z * _orientation.W));
                up.Y =
                    ((_orientation.Y * _orientation.Y) + (_orientation.W * _orientation.W)) -
                    ((_orientation.Z * _orientation.Z) + (_orientation.X * _orientation.X));
                up.Z = 2.0f *
                    ((_orientation.Y * _orientation.Z) + (_orientation.X * _orientation.W));
                return up;
            }
        }

        public Vector3 Right
        {
            get
            {
                //R v R' where v = (1,0,0,0)
                //The equation can be reduced because we know the following things:
                //  1.  We're using unit quaternions
                //  2.  The initial aspect does not change
                //The reduced form of the same equation follows
                Vector3 right = Vector3.Zero;
                right.X =
                    ((_orientation.X * _orientation.X) + (_orientation.W * _orientation.W)) -
                    ((_orientation.Z * _orientation.Z) + (_orientation.Y * _orientation.Y));
                right.Y = 2.0f *
                    ((_orientation.X * _orientation.Y) + (_orientation.Z * _orientation.W));
                right.Z = 2.0f *
                    ((_orientation.X * _orientation.Z) - (_orientation.Y * _orientation.W));

                return right;

            }
        }
        #endregion

        private float _minDistance;
        public float MinDistance
        {
            get { return _minDistance; }
            set { _minDistance = value;  }
        }

        private float _maxDistance;
        public float MaxDistance
        {
            get { return _maxDistance; }
            private set { _maxDistance = value; }
        }

        private static float GetCurrentAspectRatio()
        {
            return GameManager.Instance.Game.GraphicsDevice.Viewport.AspectRatio;
        }

        private float _angle = MathHelper.PiOver4;
        public float Angle
        {
            get { return _angle; }
            set { _angle = value; RecalculateProjection(); }
        }

        public Camera(Vector3 position, Vector3 target, float minDistance, float maxDistance)
        {
            ID = Guid.NewGuid();
            Position = position;
            Direction = target - position;
            
            MinDistance = minDistance;
            MaxDistance = maxDistance;
            PositionChanged += new EventHandler<PositionChangedEventArgs>(Camera_Moved);
            DirectionChanged += new EventHandler<PositionChangedEventArgs>(Camera_Moved);
            RecalculateAll();
        }

        public Camera(Vector3 position, Vector3 target) : this(position, target,  0.1f, 3000.0f) { }

        private void RecalculateView()
        {
            View = Matrix.CreateLookAt(Position, Position + Direction, Up);
        }

        private void RecalculateProjection()
        {
            Projection = Matrix.CreatePerspectiveFieldOfView(Angle, GetCurrentAspectRatio(), MinDistance, MaxDistance);
        }

        private void RecalculateAll()
        {
            RecalculateView();
            RecalculateProjection();
        }
        

        void Camera_Moved(object sender, PositionChangedEventArgs e)
        {
            RecalculateView();
        }
    }
}
