﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Collections;
using Microsoft.Xna.Framework;

namespace GFramework.Environment.LOD
{
	class TerrainTriangle
	{
		private int lIndex, rIndex, tIndex;
		private Vector3 lPos;
		private Vector3 vCenter;
	    private Vector3 rPos, tPos;

		private TerrainTriangle lChildTriangle, rChildTriangle;
		private TerrainTriangle lNeigh, rNeigh, bNeigh;
		private TerrainTriangle parent;

		private bool _split;
		private bool Split
		{
			get { return _split; }
			set
			{
				if (value)
					addedToMergeList = false;
				_split = value;
			}
		}
		private bool addedToMergeList = false;
		private Vector2 center;

		public TerrainTriangle(TerrainTriangle parent, Vector2 tPoint, Vector2 lPoint, Vector2 rPoint, float[,] heightData)
		{
			this.parent = parent;
			int resolution = heightData.GetLength(0);


			tIndex = ((int)tPoint.X + (int)tPoint.Y * resolution);
			lIndex = ((int)lPoint.X + (int)lPoint.Y * resolution);
			rIndex = ((int)rPoint.X + (int)rPoint.Y * resolution);

		    lPos = Vector3FromIndices(lPoint, heightData);
			

			center = (lPoint + rPoint) / 2;
		    vCenter = Vector3FromIndices(center, heightData);
		    rPos = Vector3FromIndices(rPoint, heightData);
            tPos = Vector3FromIndices(tPoint, heightData);
			

			if (Vector2.Distance(lPoint, rPoint) > 1)
			{
				lChildTriangle = new TerrainTriangle(this, center, tPoint, lPoint, heightData);
				rChildTriangle = new TerrainTriangle(this, center, rPoint, tPoint, heightData);
			}
		}

        static Vector3 Vector3FromIndices(Vector2 indxs, float[,] height)
        {
            Vector3 reslt = new Vector3()
                                {
                                    X = indxs.X*Units.HeightMapUnit,
                                    Y = height[(int) indxs.X, (int) indxs.Y],
                                    Z = indxs.Y*Units.HeightMapUnit
                                };
            return reslt;
        }

		public void SetNeighbours(TerrainTriangle left, TerrainTriangle right, TerrainTriangle bottom)
		{
			lNeigh = left;
			rNeigh = right;
			bNeigh = bottom;
			UpdateChildrenNeighbours();
		}

		void UpdateChildrenNeighbours()
		{
			if (lChildTriangle != null)
			{
				TerrainTriangle bNeightRightChild = null;
				TerrainTriangle bNeightLeftChild = null;
				TerrainTriangle lNeightRightChild = null;
				TerrainTriangle rNeightLeftChild = null;

				if (bNeigh != null)
				{
					bNeightLeftChild = bNeigh.lChildTriangle;
					bNeightRightChild = bNeigh.rChildTriangle;
				}

				if (lNeigh != null)
				{
					lNeightRightChild = lNeigh.rChildTriangle;
				}

				if (rNeigh != null)
				{
					rNeightLeftChild = rNeigh.lChildTriangle;
				}

				lChildTriangle.SetNeighbours(rChildTriangle, bNeightRightChild, lNeightRightChild);
				rChildTriangle.SetNeighbours(bNeightLeftChild, lChildTriangle, rNeightLeftChild);
			}
		}

		public void AddIndices(IList<int> indices)
		{
			indices.Add(lIndex);
			indices.Add(tIndex);
			indices.Add(rIndex);
		}

		private void PropagateSplitList(ConcurrentStack<TerrainTriangle> splitList)
		{
			if (!Split)
			{
				Split = true;
				splitList.Push(this);
				if (bNeigh != null)
					bNeigh.PropagateSplitList(splitList);
				if (parent != null)
					parent.PropagateSplitList(splitList);
			}
		}

		public bool ShouldSplit(Matrix mvp, BoundingFrustum frustum)
		{
			bool result = false;
			if (frustum.Contains(vCenter) != ContainmentType.Disjoint 
            || frustum.Contains(lPos) != ContainmentType.Disjoint
                || frustum.Contains(rPos) != ContainmentType.Disjoint
                || frustum.Contains(tPos) != ContainmentType.Disjoint)
			{
				Vector4 lScreenPos = Vector4.Transform(lPos, mvp);
				Vector4 aScreenPos = Vector4.Transform(vCenter, mvp);
				lScreenPos /= lScreenPos.W;
				aScreenPos /= aScreenPos.W;

				Vector4 difference = lScreenPos - aScreenPos;
				Vector2 aScreenDifference = new Vector2(difference.X, difference.Y);

				float eps = 0.1f;
				if (aScreenDifference.Length() > eps)
				{
					result = true;
				}
			}
			return result;
		}

		private bool CanBeSplitted
		{
			get
			{
				return lChildTriangle != null && !Split;
			}
		}

		public void FillSplitList(ConcurrentStack<TerrainTriangle> splitList, ConcurrentStack<TerrainTriangle> remainders, Matrix wvp, BoundingFrustum frustum)
		{
			if (CanBeSplitted && ShouldSplit(wvp, frustum))
			{
				PropagateSplitList(splitList);
				return;
			}
			
			remainders.Push(this);
		}

		private bool CanMerge
		{
			get
			{
				bool canNotMerge = false;
				if (lChildTriangle != null)
					canNotMerge |= lChildTriangle.Split;
				if (rChildTriangle != null)
					canNotMerge |= rChildTriangle.Split;
				if (bNeigh != null)
				{
					if (bNeigh.lChildTriangle != null)
					{
						canNotMerge |= bNeigh.lChildTriangle.Split;
					}

					if (bNeigh.rChildTriangle != null)
					{
						canNotMerge |= bNeigh.rChildTriangle.Split;
					}
				}
				return !canNotMerge;
			}
		}

		public void ProcessSplitList(List<TerrainTriangle> drawList)
		{
			if (!rChildTriangle.Split)
				drawList.Add(rChildTriangle);
			if (!lChildTriangle.Split)
				drawList.Add(lChildTriangle);
		}

		public void CreateMergeList(ConcurrentStack<TerrainTriangle> mergeList, ConcurrentStack<TerrainTriangle> leftOverList, Matrix wvp, BoundingFrustum frustum)
		{
			bool canNotMerge = false;
			if (parent != null)
			{
				canNotMerge = !parent.CheckMerge(mergeList, wvp, frustum);
			}

			if (canNotMerge)
				leftOverList.Push(this);
		}

		private bool CheckMerge(ConcurrentStack<TerrainTriangle> mergeList, Matrix wvp, BoundingFrustum frustum)
		{
			bool shouldMerge = false;
			if (!addedToMergeList)
			{
				if (CanMerge)
				{
					if (!ShouldSplit(wvp, frustum))
						shouldMerge = true;
					if (bNeigh != null)
					{
						if (bNeigh.ShouldSplit(wvp, frustum))
							shouldMerge = false;
					}
				}
			}
			if (shouldMerge)
			{
				addedToMergeList = true;
				mergeList.Push(this);
				if (bNeigh != null)
				{
					mergeList.Push(bNeigh);
				}
			}
			return addedToMergeList;
		}

		public void ProcessMergeList(List<TerrainTriangle> drawTriangles, Matrix wvp, BoundingFrustum frustum)
		{
			Split = false;
			drawTriangles.Add(this);
		}

		public void ProcessRemainders(List<TerrainTriangle> newTriangleList)
		{
			if (!Split)
				newTriangleList.Add(this);
		}
	}
}
