﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Collections;
using GFramework.Base;
using GFramework.Environment.LOD;
using GFramework.UI;
using GFramework.Vertices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace GFramework.Environment
{
    public class Terrain
    {
        private Texture2D _heightTexture;

        private VertexBuffer _vertexBuffer;

        private List<TerrainTriangle> _triangleList;
        private List<int> _indexList;
        private ListConnectedIndexBuffer<int> _cBuffer;
    	private bool camSet = false;

        public Terrain(Texture2D tex)
        {
            _heightTexture = tex;
            var heighData = GetHeight();
            FillBuffers(heighData);
            _triangleList = new List<TerrainTriangle>(TerrainSize/20);
            TerrainTriangle leftTriangle = new TerrainTriangle(null, new Vector2(TerrainSize-1,0), new Vector2(0,0), new Vector2(TerrainSize-1,TerrainSize-1), heighData );
                //new TerrainTriangle(null, new Vector2(0, 0), new Vector2(TerrainSize - 1, 0), new Vector2(0, TerrainSize - 1), heighData);
            TerrainTriangle rightTriangle = 
                new TerrainTriangle(null,new Vector2(0,TerrainSize-1), new Vector2(TerrainSize-1,TerrainSize-1), new Vector2(0,0), heighData );
                //new TerrainTriangle(null, new Vector2(TerrainSize-1, TerrainSize-1), new Vector2(0, TerrainSize-1), new Vector2(TerrainSize-1,0), heighData );

            leftTriangle.SetNeighbours(null,null,rightTriangle);
            rightTriangle.SetNeighbours(null, null, leftTriangle);

            _triangleList.Add(leftTriangle);
            _triangleList.Add(rightTriangle);

            _indexList = new List<int>();
            foreach (TerrainTriangle tr in _triangleList)
            {
                tr.AddIndices(_indexList);
            }

            _cBuffer = new ListConnectedIndexBuffer<int>(_indexList,GameManager.Instance.Game.GraphicsDevice);
			
        }

    	public int TrianglesCount
    	{
    		get
    		{
    			return _triangleList.Count;
    		}
    	}

    	private bool shouldRecalculate = true;
		void Current_PositionChanged(object sender, PositionChangedEventArgs e)
		{
			shouldRecalculate = true;
		}

        private int TerrainSize
        {
            get {  return 513; }
           // return  _heightTexture.Width; }
        }
        
        private float SmoothHeight(float rise)
        {
            return 0.25f*rise*rise;
        }

        private float[,] GetHeight()
        {
            Color[] colors = new Color[_heightTexture.Width * _heightTexture.Height];
            float[,] heights = new float[TerrainSize,TerrainSize];
            _heightTexture.GetData(colors);
            for(int i=0; i< TerrainSize; i++)
                for(int j=0; j< TerrainSize;j++)
                {
                    Color clr = colors[i + j*_heightTexture.Width];
                    heights[i, j] = clr.R; //SmoothHeight( clr.R );
                }
            return heights;
        }

        private void FillBuffers(float[,] height)
        {
            VertexPositionColor[] vertex = new VertexPositionColor[TerrainSize * TerrainSize];
            for(int i=0; i< TerrainSize; i++)
                for (int j = 0; j < TerrainSize; j++)
                {
                    Vector3 position = new Vector3(i * Units.HeightMapUnit, height[i, j], j * Units.HeightMapUnit);
                    vertex[i + j * TerrainSize] = new VertexPositionColor(position, Color.White);
                }

            _vertexBuffer = new VertexBuffer(GameManager.Instance.Game.GraphicsDevice, typeof(VertexPositionColor),vertex.Length, BufferUsage.WriteOnly);
            _vertexBuffer.SetData(vertex);
        }

        public Matrix MiddleTranslation
        {
            get
            {
                //return Matrix.Identity;
                return Matrix.CreateTranslation(-TerrainSize*Units.HeightMapUnit/2, 0, -TerrainSize*Units.HeightMapUnit/2);
            }
        }

		//List<TerrainTriangle> splitList = new List<TerrainTriangle>();
		

		ConcurrentStack<TerrainTriangle> splitList = new ConcurrentStack<TerrainTriangle>();
		ConcurrentStack<TerrainTriangle> mergeList = new ConcurrentStack<TerrainTriangle>();
		ConcurrentStack<TerrainTriangle> remainderList = new ConcurrentStack<TerrainTriangle>();
		ConcurrentStack<TerrainTriangle> leftOverList = new ConcurrentStack<TerrainTriangle>();

        private void UpdateTriangles(ICamera camera)
        {
			if (!camSet)
			{
				GameManager.Instance.CameraManager.Current.PositionChanged += new EventHandler<PositionChangedEventArgs>(Current_PositionChanged);
				camSet = true;
			}

        	
			splitList.Clear();
			mergeList.Clear();
			remainderList.Clear();
			leftOverList.Clear();
			
			List<TerrainTriangle> newTriangleList = new List<TerrainTriangle>();

            Matrix wvp = MiddleTranslation * camera.View * camera.Projection;
            BoundingFrustum frustum = new BoundingFrustum(wvp);
			
			Parallel.ForEach(_triangleList, triangle=>triangle.FillSplitList(splitList, remainderList,wvp, frustum));
			
            foreach (TerrainTriangle triangle in splitList)
            {
                triangle.ProcessSplitList(newTriangleList);
            }

			Parallel.ForEach(remainderList, triangle=>triangle.CreateMergeList(mergeList, leftOverList,wvp, frustum));
			

            foreach (TerrainTriangle triangle in mergeList)
            {
                triangle.ProcessMergeList(newTriangleList, wvp, frustum);
            }

            foreach (TerrainTriangle triangle in leftOverList)
            {
                triangle.ProcessRemainders(newTriangleList);
            }

            if (newTriangleList.Count > 0)
            {
                _triangleList = newTriangleList;
                _triangleList.TrimExcess();
            	UserInterface.UI.TrianglesCount = TrianglesCount;
            }
        }

        private void UpdateIndexBuffer()
        {
            _indexList.Clear();
            if (_triangleList.Count > 0)
            {
                foreach (TerrainTriangle triangle in _triangleList)
                {
                    triangle.AddIndices(_indexList);
                }

                _cBuffer.UpdateBuffer();
            }
        }

        public void Draw(GameTime gameTime, ICamera camera )
        {
            if (_triangleList.Count > 0)
            {
                var device = GameManager.Instance.Game.GraphicsDevice;
                BasicEffect eff = new BasicEffect(device, null);

                eff.View = camera.View;
                eff.World = MiddleTranslation;
                eff.Projection = camera.Projection;
                device.Indices = _cBuffer.Buffer; //_indexBuffer;
                device.VertexDeclaration = new VertexDeclaration(device, VertexPositionColor.VertexElements);
                device.Vertices[0].SetSource(_vertexBuffer, 0, VertexPositionColor.SizeInBytes);

                eff.Begin();
                foreach (EffectPass pass in eff.CurrentTechnique.Passes)
                {
                    pass.Begin();
                    device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, TerrainSize*TerrainSize, 0, _triangleList.Count);
                    pass.End();
                }
                eff.End();
                device.Indices = null;
            }
        }

        public void Update(GameTime gameTime)
        {
			//if (shouldRecalculate)
			{
				UpdateTriangles(GameManager.Instance.CameraManager.Current);
				shouldRecalculate = false;
			}
        	UpdateIndexBuffer();
        }

    }
}
