﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace GFramework.Input
{
    [Flags]
    public enum MouseButtons { None = 0, Left = 1, Middle = 2, Right = 4 }

    public class MouseIO
    {
        public Texture2D CurrentCursor { get; set; }
        private Color _cursorColor = Color.White;

        public event EventHandler<MouseButtonEventArgs> MouseButtonsPressed;
        public event EventHandler<MouseButtonEventArgs> MouseButtonsReleased;
        public event EventHandler<MouseScrolledEventArgs> MouseWheelScrolled;

        public event EventHandler<AllMouseChanges> DragStarted;
        public event EventHandler<AllMouseChanges> DragEnded;
        public event EventHandler<AllMouseChanges> DragContinued;

        public event EventHandler<MouseMovedEventArgs> MouseMoved;


        private MouseState _oldState, _currentState;
        private bool _dragState;

        public void Update()
        {
            _oldState = _currentState;
            _currentState = Mouse.GetState();

            Vector2 move = new Vector2(_currentState.X - _oldState.X, _currentState.Y - _oldState.Y);
            PointPosition = new Point(_currentState.X, _currentState.Y);
            VectorPosition = new Vector2(_currentState.X, _currentState.Y);

            bool lPressed = _oldState.LeftButton == ButtonState.Released && _currentState.LeftButton == ButtonState.Pressed;
            bool rPressed = _oldState.RightButton == ButtonState.Released && _currentState.RightButton == ButtonState.Pressed;
            bool mPressed = _oldState.MiddleButton == ButtonState.Released && _currentState.MiddleButton == ButtonState.Pressed;

            bool lReleased = _oldState.LeftButton == ButtonState.Pressed && _currentState.LeftButton == ButtonState.Released;
            bool rReleased = _oldState.RightButton == ButtonState.Pressed && _currentState.RightButton == ButtonState.Released;
            bool mReleased = _oldState.MiddleButton == ButtonState.Pressed && _currentState.MiddleButton == ButtonState.Released;

            int scrollDelta = _currentState.ScrollWheelValue - _oldState.ScrollWheelValue;
            bool scrolled = scrollDelta != 0;

            bool moved = move != Vector2.Zero;
            bool dragged = _oldState.LeftButton == ButtonState.Pressed && _currentState.LeftButton == ButtonState.Pressed;
            if (moved)
                LastMove = move;

            AllMouseChanges args = new AllMouseChanges(this);
            RaiseDragEvents(dragged, args);

            _dragState = dragged;

            RaiseReleased(mReleased, rReleased, lReleased);
            RaisePressed(mPressed, rPressed, lPressed);
            if (scrolled)
                RaiseScrolled(scrollDelta);

            RaiseMouseMoved(move, moved);
        }

        private void RaiseDragEvents(bool dragged, AllMouseChanges args)
        {
            if (dragged && !_dragState)
            {
                if (DragStarted != null)
                {
                    DragStarted(this, args);
                }
            }

            if (_dragState && !dragged)
            {
                if (DragEnded != null)
                {
                    DragEnded(this, args);
                }
            }

            if (_dragState && dragged)
            {
                if (DragContinued != null)
                {
                    DragContinued(this, args);
                }
            }
        }

        private void RaiseMouseMoved(Vector2 move, bool moved)
        {
            if (moved)
            {
                if (MouseMoved != null)
                {
                    MouseMovedEventArgs args = new MouseMovedEventArgs(new Point(_oldState.X, _oldState.Y), PointPosition, move, this);
                    MouseMoved(this, args);
                }
            }
        }

        private void RaiseScrolled(int delta)
        {
            if (MouseWheelScrolled != null)
            {
                MouseScrolledEventArgs args = new MouseScrolledEventArgs(delta, this);
                MouseWheelScrolled(this, args);
            }
        }

        private void RaiseReleased(bool mReleased, bool rReleased, bool lReleased)
        {
            if (mReleased || rReleased || lReleased)
            {
                if (MouseButtonsReleased != null)
                {
                    MouseButtons buttons = MouseButtons.None;
                    if (lReleased)
                        buttons = buttons | MouseButtons.Left;
                    if (mReleased)
                        buttons = buttons | MouseButtons.Middle;
                    if (rReleased)
                        buttons = buttons | MouseButtons.Right;
                    MouseButtonEventArgs args = new MouseButtonEventArgs(buttons, this);
                    MouseButtonsReleased(this, args);
                }
            }
        }

        private void RaisePressed(bool mPressed, bool rPressed, bool lPressed)
        {
            if (mPressed || rPressed || lPressed)
            {
                if (MouseButtonsPressed != null)
                {
                    MouseButtons buttons = MouseButtons.None;
                    if (lPressed)
                        buttons = buttons | MouseButtons.Left;
                    if (mPressed)
                        buttons = buttons | MouseButtons.Middle;
                    if (rPressed)
                        buttons = buttons | MouseButtons.Right;
                    MouseButtonEventArgs args = new MouseButtonEventArgs(buttons, this);
                    MouseButtonsPressed(this, args);
                }
            }
        }

        public Point PointPosition { get; private set; }
        public Vector2 VectorPosition { get; private set; }
        private bool _drawMouse = true;
        public bool DrawMouse
        {
            get { return _drawMouse; }
            set { _drawMouse = value; }
        }

        public Vector2 LastMove { get; private set; }

        public void Draw(SpriteBatch sBatch)
        {
            if (DrawMouse)
                sBatch.Draw(CurrentCursor, VectorPosition, _cursorColor);
        }
    }

    public class MouseButtonEventArgs : EventArgs
    {
        public MouseButtons Buttons { get; private set; }
        public MouseIO Manager { get; private set; }

        public MouseButtonEventArgs(MouseButtons buttons, MouseIO mouseIo)
        {
            Buttons = buttons;
            Manager = mouseIo;
        }
    }

    public class MouseScrolledEventArgs : EventArgs
    {
        public int Delta { get; private set; }
        public MouseIO Manager { get; private set; }

        public MouseScrolledEventArgs(int delta, MouseIO mouse)
        {
            Delta = delta;
            Manager = mouse;
        }
    }

    public class AllMouseChanges : EventArgs
    {
        public MouseIO Manager { get; private set; }

        public AllMouseChanges(MouseIO mouse)
        {
            Manager = mouse;
        }
    }

    public class MouseMovedEventArgs : EventArgs
    {
        public MouseIO Manager { get; private set; }
        public Point PrevPosition { get; private set; }
        public Point CurrentPosition { get; private set; }
        public Vector2 MovedDelta { get; private set; }

        public MouseMovedEventArgs(Point prev, Point curr, Vector2 delta, MouseIO mouse)
        {
            Manager = mouse;
            PrevPosition = prev;
            CurrentPosition = curr;
            MovedDelta = delta;
        }
    }

}
